<?php

namespace Nest\Events;

use Nest\Exceptions\InvalidTypeException;
use function Nest\Utils\expect_type;
use function Nest\Utils\get_all_classes;

/**
 * Events manager class.
 */
class EventsManager
{
	/**
	 * Events listeners.
	 * If false is returned, stop event propagation.
	 * @var array<class-string, (callable(Event $event): (void|bool))[]>
	 */
	protected array $listeners = [];

	/**
	 * Add an event listener to the listeners list.
	 * @template T of Event
	 * @param class-string<T> $eventType The event type class to listen for.
	 * @param callable(T $event): (void|bool) $listener The event listener to add. If false is returned, stop event propagation.
	 * @return void
	 * @throws InvalidTypeException
	 */
	public function addListener(string $eventType, callable $listener): void
	{
		expect_type($eventType, Event::class);

		// Initialize the listeners list if it's empty.
		if (empty($this->listeners[$eventType])) $this->listeners[$eventType] = [];
		// Add the listener.
		$this->listeners[$eventType][] = $listener;
	}

	/**
	 * Remove the given listener from the listeners list.
	 * @template T of Event
	 * @param class-string<T> $eventType The event type class.
	 * @param callable(T $event): (void|bool) $listener The listener to remove.
	 * @return void
	 * @throws InvalidTypeException
	 */
	public function removeListener(string $eventType, callable $listener): void
	{
		expect_type($eventType, Event::class);

		if (!empty($this->listeners[$eventType]))
		{
			foreach ($this->listeners[$eventType] as $key => $currentListener)
			{ // For each listeners, if it's matching the given one, remove it.
				if ($currentListener == $listener) unset($this->listeners[$eventType][$key]);
			}
		}
	}

	/**
	 * Fire an event.
	 * @param Event $event The event to fire.
	 * @return void
	 */
	public function fire(Event $event): void
	{
		foreach (get_all_classes($event) as $className)
		{ // For each class, from the most specific to the least specific.
			if (!empty($this->listeners[$className]))
			{ // If there are listeners for the current class.
				foreach ($this->listeners[$className] as $listener)
				{ // Call each listener.
					$result = $listener($event);
					if ($result === false)
						// If the result is false, stop firing.
						return;
				}
			}
		}
	}
}
